
(** Check if 2^n <= (n+1)^d by taking log of both sides *)
let small_n (d : int) (n : int) : bool =
  let d1 = float_of_int d in
  let n1 = float_of_int n in
    n1 *. log 2.0 <= d1 *. log (n1 +. 1.0)

(** compute the largest n such that 2^n <= (n+1)^d   *)
let n_over_logn_inv (d : int) : int = 
  let rec search a b =
    (* pre "a <= b & 
            small_n d a";
    *)
    if b <= a + 1 then
      if small_n d b then b 
      else a
    else 
      let c = (a + b) / 2 in
	if small_n d c then search c b
	else search a c
  in
  let d1 = float_of_int d in
  let a0 = d1 *. log d1 in
  let b0 = 2.0 *. d1 *. (1.0 +. log d1) in
    search (int_of_float a0) (int_of_float b0)

let find_sparseness_bound (d : int) =
  if d <= 3 then d
  else 
    n_over_logn_inv d
